#! /usr/bin/python3

# Read SPLConqueror measuremnts file, add random noise to the values, and write
# out one ore more measuremnts files with noise in the values.
#
# The files with noised measurements are used to test the stability of the
# learned models in the presence of measurement errors (simulated by the
# noise).
#
# Usage: ./add_noise_to_measurements.py

import os, sys
import lxml.etree as ET
import numpy as np

# Parameteres for generating random samples form normal distribution
# (numpy.random.normal).
LOC = 0
RELATIVE_SCALE = 0.75

MEASUREMENTS_PATHNAME = "all_measurements.xml"
NOISED_MEASUREMENTS_PATHNAME_PATTERN = "all_measurements_{0}.xml"
NUM_GENERATION_ROUNDS = 6

def write_noised_measurements(all_noised_measurents):
    for relative_scale, noised_measurements in all_noised_measurents:
        out_pathname = NOISED_MEASUREMENTS_PATHNAME_PATTERN.format(relative_scale)
        el_root = ET.Element('results')
        for configuration, noised_value in noised_measurements:
            el_row = ET.SubElement(el_root, 'row')
            el_data_configuration = ET.SubElement(el_row, 'data', columname='Configuration')
            el_data_configuration.text = configuration
            el_data_noised_value = ET.SubElement(el_row, 'data', columname='Measured Value')
            el_data_noised_value.text = "{0:.2f}".format(noised_value)
        # print(ET.tostring(el_root, encoding='unicode', pretty_print=True))
        tree = ET.ElementTree(element=el_root)
        tree.write(out_pathname, pretty_print=True)

def add_noise(measurements, loc, relative_scale):
    noised_measurements = []
    for configuration, value in measurements:
        scale = value / 100 * relative_scale
        noise = np.random.normal(loc=loc, scale=scale)
        noised_value = value + noise
        noised_measurements.append((configuration, noised_value))
        # print(relative_scale, "\t", "{0:.2f}".format(noise), "\t", "{0:.2f}".format(value), "\t", "{0:.2f}".format(noised_value))
    return noised_measurements

def parse_measurements():
    ''' Parse a mesurement file.  Return a list of typels: (configuration, measured_value).
    '''
    measurements = []
    tree = ET.parse(MEASUREMENTS_PATHNAME)
    root = tree.getroot()
    for measurement in root.findall('row'):
        configuration = measurement[0].text.strip()
        value = float(measurement[1].text.replace(',', '.'))
        measurements.append((configuration, value))
    return measurements

def main():
    measurements = parse_measurements()
    all_noised_measurents = []
    current_relative_scale = RELATIVE_SCALE
    for i in range(NUM_GENERATION_ROUNDS):
        noised_measurements = add_noise(measurements, LOC, current_relative_scale)
        all_noised_measurents.append((current_relative_scale, noised_measurements))
        current_relative_scale *= 2
    write_noised_measurements(all_noised_measurents)

if __name__ == '__main__':
    main()
