#!/usr/bin/python3

import sys, os
import glob
import plot_interaction_log
import plot_all_measurement
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages

CASE_STUDIES = [
    'sanity_check/ajstats_noise',
    'sanity_check/apache_noise',
    'sanity_check/berkeleydbc_noise',
    'sanity_check/berkeleydbj_noise',
    'sanity_check/clasp_noise',
    # 'sanity_check/dune_noise',
    # 'sanity_check/hsmgp_noise',
    'sanity_check/llvm_noise',
    'sanity_check/lrzip_noise',
    'sanity_check/x264_noise',
]

NOISE_SUFFIX = ['00.00.log', '00.75.log', '01.50.log', '03.00.log', '06.00.log', '12.00.log', '24.00.log']
NOISE_VALUES = [0.0, 0.75, 1.5, 3.0, 6.0, 12.0, 24.0]

COLORS = ['b', 'g', 'r', 'c', 'm', 'y', 'k',]
CURRENT_COLOR = 'b'
def reset_current_color():
    global CURRENT_COLOR
    CURRENT_COLOR = 'b'
def next_current_color():
    if 'color_counter' not in next_current_color.__dict__:
        next_current_color.color_counter = 0
    else:
        next_current_color.color_counter += 1
    global CURRENT_COLOR
    CURRENT_COLOR = COLORS[next_current_color.color_counter % len(COLORS)]

def plot_cpu_time(case_study_id, learner_variant, logs):
    cpu_times = [log.final_model_log.time for log in logs]
    nonoise_time = cpu_times[0]
    relative_cpu_times = [t/nonoise_time for t in cpu_times]
    ax_cpu_time, *_ = plt.plot(NOISE_VALUES, relative_cpu_times, CURRENT_COLOR+'s--')
    plt.annotate(case_study_id, xy=(NOISE_VALUES[-1], relative_cpu_times[-1]), horizontalalignment='right', fontsize='xx-small')
    plt.ylabel("relative cpu time")
    # ymax = ((plt.ylim()[1] - plt.ylim()[0]) * 0.05) + plt.ylim()[1]
    # plt.ylim(ymin=-2, ymax=ymax)
    plt.ylim(ymin=0, ymax=18)
    #plt.ylim(ymin=-2.5, ymax=302.5)
    # xmax = ((plt.xlim()[1] - plt.xlim()[0]) * 0.05) + plt.xlim()[1]
    # plt.xlim(xmin=-0.25, xmax=xmax)
    plt.xlim(xmin=-0.25)
    plt.xlabel('noise strength (%)')
    plt.xticks(NOISE_VALUES, rotation='vertical')

    #plt.legend(ax_cpu_time, 'cpu_time', loc='lower right')
    plt.grid(True)
    plt.title('noise effect on learning time (' + learner_variant + ')')

def plot_complexity(case_study_id, learner_variant, logs):
    complexities = [log.final_model_log.complexity for log in logs]
    # nonoise_complexity = complexities[0]
    # complexitiy_increments = [c - nonoise_complexity for c in complexities]
    # ax_cpu_time, *_ = plt.plot(NOISE_VALUES, complexitiy_increments, CURRENT_COLOR+'s-')
    # plt.annotate(case_study_id, xy=(NOISE_VALUES[-1], complexitiy_increments[-1]), horizontalalignment='right', fontsize='xx-small')

    ax_cpu_time, *_ = plt.plot(NOISE_VALUES, complexities, CURRENT_COLOR+'s-')
    plt.annotate(case_study_id, xy=(NOISE_VALUES[-1], complexities[-1]), horizontalalignment='right', fontsize='xx-small')
    plt.ylabel("model complexity increment (points)")
    plt.ylim(ymin=-10, ymax=260)
    # xmax = ((plt.xlim()[1] - plt.xlim()[0]) * 0.05) + plt.xlim()[1]
    # plt.xlim(xmin=-0.25, xmax=xmax)
    plt.xlim(xmin=-0.25)
    plt.xlabel('noise strength (%)')
    plt.xticks(NOISE_VALUES, rotation='vertical')

    #plt.legend(ax_cpu_time, 'cpu_time', loc='lower right')
    plt.grid(True)
    plt.title('noise effect on model complexity (' + learner_variant + ')')

def plot_structure_error(case_study_id, learner_variant, logs):
    structure_errors = []
    original_model_combinations = [m.combination
                                   for m in logs[0].final_model_log.model]
    for log in logs:
        noise_model_combinations = [m.combination
                                    for m in log.final_model_log.model]
        union = set(original_model_combinations) | set(noise_model_combinations)
        intersect = set(original_model_combinations) & set(noise_model_combinations)
        jaccard_index = len(intersect) / len(union)
        structure_error = (1 - jaccard_index) * 100
        structure_errors.append(structure_error)
    ax_similarity, *_ = plt.plot(NOISE_VALUES, structure_errors, CURRENT_COLOR+'s--')
    plt.annotate(case_study_id, xy=(NOISE_VALUES[-1], structure_errors[-1]), horizontalalignment='right', fontsize='xx-small')
    plt.ylabel("model structure error (%)")
    # ymax = ((plt.ylim()[1] - plt.ylim()[0]) * 0.05) + plt.ylim()[1]
    # plt.ylim(ymin=-2, ymax=ymax)
    plt.ylim(ymin=-2, ymax=100)
    # xmax = ((plt.xlim()[1] - plt.xlim()[0]) * 0.05) + plt.xlim()[1]
    # plt.xlim(xmin=-0.25, xmax=xmax)
    plt.xlim(xmin=-0.25)
    plt.xlabel('noise strength (%)')
    plt.xticks(NOISE_VALUES, rotation='vertical')

    #plt.legend(ax_similarity, 'structure error', loc='lower right')
    plt.grid(True)
    plt.title('noise effect on model structure (' + learner_variant + ')')

def plot_prediction_error(case_study_id, learner_variant, logs, nonoise_measurements):
    avg_relative_errors = []
    for log in logs:
        model_log = log.final_model_log
        abs_relative_errors = []
        for configuration_string, measured_value in nonoise_measurements:
            predicted_value = model_log.predict(configuration_string.split(','))
            error = abs(measured_value - predicted_value)/measured_value * 100
            abs_relative_errors.append(error)
        avg_relative_errors.append(np.mean(abs_relative_errors))
    ax_error, *_ = plt.plot(NOISE_VALUES, avg_relative_errors, CURRENT_COLOR+'o-')
    plt.annotate(case_study_id, xy=(NOISE_VALUES[-1], avg_relative_errors[-1]), horizontalalignment='right', fontsize='xx-small')
    plt.ylabel('average perdiction error (%)')
    # ymax = ((plt.ylim()[1] - plt.ylim()[0]) * 0.05) + plt.ylim()[1]
    # plt.ylim(ymin=-0.25, ymax=ymax)
    plt.ylim(ymin=-1.2, ymax=60)
    # xmax = ((plt.xlim()[1] - plt.xlim()[0]) * 0.05) + plt.xlim()[1]
    # plt.xlim(xmin=-0.25, xmax=xmax)
    plt.xlim(xmin=-0.25)
    plt.xlabel('noise strength (%)')
    plt.xticks(NOISE_VALUES, rotation='vertical')
    #plt.legend(ax_error, 'prediction error', loc='lower right')
    plt.grid(True)
    plt.title('noise effect on prediction accuracy (' + learner_variant + ')')

def sort_logs(logs):
    # We assume that the logs are named in such a way that, after sorting them
    # alphabetically, the logs are grouped according to the learner's
    # paramteres (i.e., hierarchical, non-hierarchical, with penalty, without
    # penalty), and each group is sorted by the noise value in ascending order.
    logs = sorted(logs, key=lambda l: l.log_id)
    logs_h, logs_h_penalty, logs_nh, logs_nh_penalty = [logs[x:x+len(NOISE_SUFFIX)] for x in range(0, len(logs), len(NOISE_SUFFIX))]
    # Test the assumption about the sorted logs.
    def has_correct_suffixes(logs):
        for log_name, suffix in zip([l.log_id for l in logs], NOISE_SUFFIX):
            if not log_name.endswith(suffix):
                raise AssertionError("Can't sort logs! Wrong noise values: " + log_name + " != " + suffix)
    for logs_group in [logs_h, logs_h_penalty, logs_nh, logs_nh_penalty]:
        has_correct_suffixes(logs_group)
    if logs_h and not (logs_h[0].settings.is_hierarchical_learning and logs_h[0].settings.candidate_size_penalty == 0):
        raise AssertionError("Can't sort logs! Logs (hierarchical, no penalty):" + [logs_h])
    if logs_h_penalty and not (logs_h_penalty[0].settings.is_hierarchical_learning and logs_h_penalty[0].settings.candidate_size_penalty != 0):
        raise AssertionError("Can't sort logs! Logs (hierarchical, with penalty):" + [logs_h_penalty])
    if logs_nh and not (not logs_nh[0].settings.is_hierarchical_learning and logs_nh[0].settings.candidate_size_penalty == 0):
        raise AssertionError("Can't sort logs! Logs (non-hierarchical, no penalty):" + [logs_nh])
    if logs_nh_penalty and not (not logs_nh_penalty[0].settings.is_hierarchical_learning and logs_nh_penalty[0].settings.candidate_size_penalty != 0):
        raise AssertionError("Can't sort logs! Logs (non-hierarchical, with penalty):" + [logs_nh_penalty])
    return logs_h, logs_h_penalty, logs_nh, logs_nh_penalty

def main(case_study_dirs):
    parsed_logs = {}
    parsed_nonoise_measurements = {}
    for dir in case_study_dirs:
        case_study_id =  dir.strip(os.sep).replace(os.sep, '_')
        logs = plot_interaction_log.parse_logs(dir)
        parsed_logs[case_study_id] = logs
        nonoise_measurements = plot_all_measurement.parse_measurements(dir)
        parsed_nonoise_measurements[case_study_id] = nonoise_measurements
    num_plot_rows = 4
    num_plot_columns = 4
    plt.figure(figsize=(10 * num_plot_columns, 6 * num_plot_rows))
    for case_study_id in sorted(parsed_logs):
        logs_h, logs_h_penalty, logs_nh, logs_nh_penalty = sort_logs(parsed_logs[case_study_id])
        current_plot_position = 1
        for learner_variant, logs_group in [('hierarchical, no penalty', logs_h), ('hierarchical, with penalty', logs_h_penalty), ('non-hierarchical, no penalty', logs_nh), ('non-hierarchical, with penalty', logs_nh_penalty)]:
            plt.subplot(num_plot_rows, num_plot_columns,
                        current_plot_position)
            plot_prediction_error(case_study_id, learner_variant, logs_group, parsed_nonoise_measurements[case_study_id])
            current_plot_position += 1
            plt.subplot(num_plot_rows, num_plot_columns,
                        current_plot_position)
            plot_structure_error(case_study_id, learner_variant, logs_group)
            current_plot_position += 1
            plt.subplot(num_plot_rows, num_plot_columns,
                        current_plot_position)
            plot_complexity(case_study_id, learner_variant, logs_group)
            current_plot_position += 1
            plt.subplot(num_plot_rows, num_plot_columns,
                        current_plot_position)
            plot_cpu_time(case_study_id, learner_variant, logs_group)
            current_plot_position += 1
        next_current_color()
    plt.tight_layout() # autoadjust subplots to prevent overlapping
    ##Write PDF.
    pdf_filename = 'plt_noise_effect.pdf'
    pp = PdfPages(pdf_filename)
    pp.savefig()
    pp.close()
    # plt.show()
    plt.close()


if __name__ == '__main__':
    if len(sys.argv) < 2:
        print('Plotting: ' + str(CASE_STUDIES))
        main(CASE_STUDIES)
    else:
        main(sys.argv[1:])
